/*
 * @(#)label.c	1.2 01/30/01
 *
 * Copyright (C) 1999-2001 Tak-Shing Chan.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License as
 * published by the Free Software Foundation; either version 2 of
 * the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied
 * warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE.  See the GNU General Public License for more details.
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <ctype.h>
#include <unistd.h>
#include <pwd.h>
#include <time.h>

/* Input buffer */
#define LINES   32
#define BUFSIZE 128
static char     buffer[LINES][BUFSIZE];
static int      copies;

/* Read records from input file */
/* Returns: number of lines read */
static int
readnext(FILE *fp, int rows)
{
    static int  trunc = 0;
    int         i = 0, j, k, len, sum, count;
    char        code, net[12], post[] = {
        24, 3, 5, 6, 9, 10, 12, 17, 18, 20
    };

    /* Input loop */
    copies = 1;
    while (fgets(buffer[i], BUFSIZE - 1, fp)) {

        /* Truncate lines into buffer, and discard newlines */
        len = strlen(buffer[i]) - 1;
        if (buffer[i][len] == '\n') {
            if (trunc) {
                trunc = 0;
                continue;
            } else
                buffer[i][len] = 0;
        } else {
            if (trunc)
                continue;
            else
                trunc = 1;
        }

        /* Mark buffer as plain text */
        buffer[i][BUFSIZE - 1] = 'S';

        /* Handle requests */
        if (!strcmp(buffer[i], "####")) {
            if (i)
                return i;
            else
                continue;
        }
        if (len > 2 && buffer[i][0] == '.') {

            /* Number of copies */
            if (buffer[i][1] == 'c' && buffer[i][2] == 'p') {
                if (sscanf(buffer[i] + 3, "%u", &copies) != 1 ||
                    copies < 1)
                    copies = 1;
                continue;
            }

            /* POSTNET barcodes */
            else if (buffer[i][1] == 'p' &&
                buffer[i][2] == 'n') {
                sum = count = 0;
                for (j = 3; count <= 10 && j <= len; j++)
                    if (isdigit(code = buffer[i][j]))
                        sum += (net[count++] = code - '0');

                /* Compute check digit */
                net[count] = (10 - sum % 10) % 10;

                /* Convert string to POSTNET binary */
                buffer[i][0] = '1';
                for (j = 0; j <= count; j++) {
                    code = post[(int) net[j]];
                    for (k = 5; k; code >>= 1)
                        buffer[i][j * 5 + k--] = '0' +
                            (code & 1);
                }
                buffer[i][j * 5 + 1] = '1';
                buffer[i][j * 5 + 2] = 0;

                /* Mark buffer as POSTNET code */
                buffer[i][BUFSIZE - 1] = 'P';
            }
        }
        /* Silently discards extra rows */
        if (i < rows) i++;
    }

    /* EOF */
    return i - (i == rows);
}

int
main(int argc, char *argv[])
{
    extern char     *optarg;
    extern int      optind;
    int             c, flag = 0, bolding = 0, font = 0,
                    indent = 12, width = 288, height = 96,
                    hsize = 612, vsize = 792, point = 12,
                    hmargin, vmargin, num, page = 1, rows, hpos,
                    vpos, i;
    FILE            *fp;
    struct passwd   *passwd;
    time_t          now;
    char            *roman[] = { "Helvetica", "Courier" },
                    *bold[] = {
        "Helvetica-Bold", "Courier-Bold"
    };

    /* Process arguments */
    while ((c = getopt(argc, argv, "bci:l:p:s:")) != EOF) {
        switch (c) {
        case 'b':
            bolding = 1;
            break;
        case 'c':
            font = 1;
            break;
        case 'i':
            if (sscanf(optarg, "%d", &indent) != 1)
                flag++;
            break;
        case 'l':
            if (sscanf(optarg, "%dx%d", &width, &height) != 2 ||
                width < 1 || height < 1)
                flag++;
            break;
        case 'p':
            if (sscanf(optarg, "%dx%d", &hsize, &vsize) != 2 ||
                hsize < 1 || vsize < 1)
                flag++;
            break;
        case 's':
            if (sscanf(optarg, "%d", &point) != 1 || point < 10
                || point > 12)
                flag++;
            break;
        case '?':
            flag++;
        }
    }

    /* Illegal options? */
    if (flag || argc > optind + 1) {
        fprintf(stderr, "Usage: %s [-bc] [-i indent] [-l WxH] \
[-p WxH] [-s 10|11|12] [file]\n", argv[0]);
        exit(EXIT_FAILURE);
    }

    /* Open input file */
    if (argc == optind)
        fp = stdin;
    else if (!(fp = fopen(argv[optind], "r"))) {
        perror(argv[optind]);
        exit(EXIT_FAILURE);
    }

    /* PostScript Level 2 output (DSC 3.0) */
    printf("%%!PS-Adobe-3.0\n");
    printf("%%%%Title: ");
    if (argc == optind)
        printf("stdin\n");
    else
        printf("%s\n", argv[optind]);
    if ((passwd = getpwuid(getuid())))
        printf("%%%%For: %s\n", passwd->pw_gecos);
    printf("%%%%Creator: Label\n");
    time(&now);
    printf("%%%%CreationDate: %s", ctime(&now));
    printf("%%%%Pages: (atend)\n");
    printf("%%%%Orientation: Portrait\n");
    printf("%%%%DocumentMedia: Labels %d %d 0 () ()\n", hsize,
        vsize);
    printf("%%%%DocumentNeededResources: font %s\n",
        roman[font]);
    if (bolding)
        printf("%%%%+ font %s\n", bold[font]);
    printf("%%%%LanguageLevel: 2\n");
    printf("%%%%EndComments\n");
    printf("%%%%BeginProlog\n");
    printf("%%%%EndProlog\n");
    printf("%%%%BeginSetup\n");
    printf("%%%%IncludeResource: font %s\n", roman[font]);
    if (bolding)
        printf("%%%%IncludeResource: font %s\n", bold[font]);
    hmargin = (hsize / (hsize / width) - width) / 2;
    vmargin = (vsize - ((vsize - 72) / height) * height) / 2;
    printf("/C { initclip %d %d %d rectclip ", vmargin, width -
        indent, vsize - 2 * vmargin);
    if (bolding) {
        printf("/%s %d selectfont } bind def\n", bold[font],
            point);
        printf("/R { /%s %d selectfont } bind def\n",
            roman[font], point);
    } else {
        printf("/%s %d selectfont } bind def\n", roman[font],
            point);
    }
    printf("/S { moveto show } bind def\n");
    printf("/F { 3.3 9 rmoveto 0 -9 rlineto } bind def\n");
    printf("/H { 3.3 3.6 rmoveto 0 -3.6 rlineto } bind def\n");
    printf("/P { moveto { 49 eq { F } { H } ifelse } forall \
stroke } bind def\n");
    printf("%%%%EndSetup\n");

    /* Main loop */
    if ((rows = height / (point + 1)) >= LINES)
        rows = LINES - 1;
    for (num = readnext(fp, rows); num; page++) {
        printf("%%%%Page: %d %d\n", page, page);
        printf("%%%%BeginPageSetup\n");
        printf("/pagelevel save def\n");
        printf("1.44 setlinewidth\n");
        printf("%%%%EndPageSetup\n");
        for (vpos = vsize - vmargin; num && vpos >= height +
            vmargin; vpos -= height) {
            for (hpos = hmargin; num && hpos <= hsize - hmargin -
                width; hpos += 2 * hmargin + width) {
                for (i = 0; i < num; i++) {
                    switch (i) {
                    case 0:
                        printf("%d C ", hpos);
                        break;
                    case 1:
                        if (bolding)
                            printf("R ");
                    }
                    printf("(%s) %d %d %c\n", buffer[i], hpos +
                        indent, vpos + ((point + 1) * (num - 2 *
                        i - 2) - height) / 2, buffer[i][BUFSIZE -
                        1]);
                }
                if (!--copies)
                    num = feof(fp) ? 0 : readnext(fp, rows);
            }
        }
        printf("pagelevel restore\n");
        printf("showpage\n");
    }

    /* Trailer */
    printf("%%%%Trailer\n");
    printf("%%%%Pages: %d\n", page - 1);
    printf("%%%%EOF\n");
    if (argc != optind)
        fclose(fp);
    return EXIT_SUCCESS;
}
